import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';

typedef DraggableItemBuilder<T> = Widget Function(
  BuildContext context,
  T item,
  int index,
);

class DraggableGridView<T> extends StatefulWidget {
  final List<T>? initialValue;
  final DraggableItemBuilder<T> builder;
  final Duration animateDuration;
  final GridController<T>? controller;
  final SliverGridDelegate gridDelegate;
  final DraggableConfiguration<T> draggableConfiguration;
  final ValueChanged<List<T>>? onAccept;
  final void Function(List<T>? items, int index)? onDragCompleted;

  const DraggableGridView({
    this.draggableConfiguration = const DraggableConfiguration.disabled(),
    required this.builder,
    this.initialValue,
    this.controller,
    required this.gridDelegate,
    this.animateDuration = const Duration(milliseconds: 100),
    this.onAccept,
    this.onDragCompleted,
    Key? key,
  }) : super(key: key);
  @override
  State<StatefulWidget> createState() => _DraggableGridViewState<T>();
}

class _DraggableGridViewState<T> extends State<DraggableGridView<T>> {
  GridController<T>? _controller;
  GridController<T> get _effectiveController =>
      widget.controller ?? _controller!;

  @override
  void initState() {
    super.initState();
    if (widget.controller == null) {
      _createLocalController(widget.initialValue ?? []);
    }
  }

  @override
  void dispose() {
    _controller?.dispose();
    super.dispose();
  }

  void _createLocalController(List<T> items) {
    _controller = GridController<T>(items);
  }

  @override
  void didUpdateWidget(DraggableGridView<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      if (oldWidget.controller != null && widget.controller == null) {
        _createLocalController(oldWidget.controller!.value);
      }

      if (widget.controller != null) {
        if (oldWidget.controller == null) {
          _controller!.dispose();
          _controller = null;
        }
      }
    }
  }

  @override
  Widget build(BuildContext context) {
    return OrientationBuilder(builder: (context, o) {
      return GridView.builder(
        shrinkWrap: true,
        gridDelegate: widget.gridDelegate,
        itemCount: _effectiveController.value.length,
        itemBuilder: (context, index) {
          return DraggableGridItem<T>(
            widget.animateDuration,
            _effectiveController.createIfNotExists(index),
            _effectiveController,
            widget.builder,
            _effectiveController.value[index],
            widget.draggableConfiguration,
            widget.onAccept,
            widget.onDragCompleted,
          );
        },
      );
    });
  }
}

class GridController<T> extends ValueNotifier<List<T>> {
  final Map<int, _DraggableController> _controllers = {};
  final Map<int, List<int>> _animationIndexsMap = {};
  final List<VoidCallback> _animateActions = [];
  final Map<int, VoidCallback> _completeCallback = {};
  List<T>? _animateItems;

  int _gen = 0;

  GridController(List<T> items) : super(items);

  _DraggableController createIfNotExists(int index) {
    return _controllers.putIfAbsent(index, () => _DraggableController(index));
  }

  @override
  void dispose() {
    _clearAnimations();
    for (final _DraggableController element in _controllers.values) {
      element.dispose();
    }
    _controllers.clear();
    super.dispose();
  }

  void _clearAnimations() {
    _animationIndexsMap.clear();
    _animateActions.clear();
    _completeCallback.clear();
  }

  void _drag(int start, int end) {
    _animateActions.add(() {
      _dragAnimate(start, end);
    });

    _nextAnimate();
  }

  void removeAnimate(List<int> indexes, ValueChanged<List<T>> onCompleted) {
    _animateActions.add(() {
      final List<T>? items = _getRemovedList(indexes);
      if (items == null) {
        return;
      }
      final int gen = ++_gen;
      for (final element in _controllers.values) {
        final T _item = value[element.index];
        final int index = items.indexOf(_item);
        element._currentIndex = index;
      }
      _animate(gen, () {
        onCompleted(items);
      });
    });

    _nextAnimate();
  }

  @override
  set value(List<T> newValue) {
    if (value.length == newValue.length) {
      bool equals = true;

      for (int i = 0; i < value.length; i++) {
        if (value[i] != newValue[i]) {
          equals = false;
          break;
        }
      }

      if (equals) {
        return;
      }
    }

    _clearAnimations();
    for (final _DraggableController element in _controllers.values) {
      element.dispose();
    }
    _controllers.clear();
    _animateItems = null;
    super.value = newValue;
  }

  List<T>? _getRemovedList(List<int> indexes) {
    final List<int> validIndexes = indexes
        .where((element) => element >= 0 && element < value.length)
        .toSet()
        .toList()
      ..sort((a, b) => -a.compareTo(b));
    if (validIndexes.isEmpty) {
      return null;
    }
    _animateItems ??= List.of(value);
    validIndexes
        .map((e) => _controllers[e]!.currentIndex)
        .forEach(_animateItems!.removeAt);
    return _animateItems;
  }

  void _onAnimateComplete(int gen) {
    _completeCallback.remove(gen)?.call();
  }

  /// [start] the index of DraggableItem
  ///
  /// [end] the **current index** of DraggableItem
  void _dragAnimate(int start, int end) {
    final int gen = ++_gen;
    _animateItems ??= List.of(value);
    final T item = value[start];
    _animateItems!
        .insert(end, _animateItems!.removeAt(_animateItems!.indexOf(item)));
    for (final element in _controllers.values) {
      final T _item = value[element.index];
      element._currentIndex = _animateItems!.indexOf(_item);
    }
    _animate(gen);
  }

  void _animate(int gen, [VoidCallback? onComplete]) {
    final List<int> animationIndexs = [];
    for (final _DraggableController controller in _controllers.values) {
      if (controller.currentIndex == -1) {
        animationIndexs.add(controller.index);
        controller.value = _AnimateInfo.fadeIn(gen);
      } else if (controller.index != controller.currentIndex) {
        final Offset targetPosition =
            _getPixelPosition(controller.currentIndex);
        final Offset position = _getPixelPosition(controller.index);
        animationIndexs.add(controller.index);
        controller.value = _AnimateInfo.slide(
            controller.value.end,
            Offset(targetPosition.dx - position.dx,
                targetPosition.dy - position.dy),
            gen);
      } else {
        if (controller.value.end != Offset.zero) {
          animationIndexs.add(controller.index);
          controller.value =
              _AnimateInfo.slide(controller.value.end, Offset.zero, gen);
        }
      }
    }
    if (animationIndexs.isNotEmpty) {
      _animationIndexsMap[gen] = animationIndexs;
      if (onComplete != null) {
        _completeCallback[gen] = onComplete;
      }
    } else {
      if (onComplete != null) {
        onComplete.call();
      }
    }
  }

  void _onAnimateEnd(int gen, int index) {
    final List<int>? indexes = _animationIndexsMap[gen];
    if (indexes != null) {
      indexes.remove(index);
      if (indexes.isEmpty) {
        _animationIndexsMap.remove(gen);
        _onAnimateComplete(gen);
      }
    }
    _nextAnimate();
  }

  void _nextAnimate() {
    if (_animationIndexsMap.isEmpty && _animateActions.isNotEmpty) {
      final VoidCallback animate = _animateActions.removeAt(0);
      animate();
    }
  }

  Offset _getPixelPosition(int index) {
    for (final _DraggableController controller in _controllers.values) {
      if (controller.index == index) {
        return controller.position!;
      }
    }
    throw Exception('can not get pixel position at $index');
  }
}

class DraggableGridItem<T> extends StatefulWidget {
  final _DraggableController controller;
  final GridController<T> gridController;
  final DraggableItemBuilder<T> builder;
  final Duration duration;
  final T item;
  final DraggableConfiguration<T> draggableConfiguration;

  /// called when drag canceled or accepted
  final ValueChanged<List<T>>? onAccept;

  /// called when draggable item is accepted by **OTHER** [DragTarget]
  final void Function(List<T>? items, int index)? onDragCompleted;
  DraggableGridItem(
    this.duration,
    this.controller,
    this.gridController,
    this.builder,
    this.item,
    this.draggableConfiguration,
    this.onAccept,
    this.onDragCompleted,
  ) : super(key: controller.key);

  @override
  State<StatefulWidget> createState() => _DraggableItemState<T>();
}

class _DraggableItemState<T> extends State<DraggableGridItem<T>>
    with SingleTickerProviderStateMixin {
  late final AnimationController _controller = AnimationController(
    duration: widget.duration,
    vsync: this,
  );

  final ValueNotifier<Size?> _sizeNotifier = _QuietlySizeNotifier(null);
  late _AnimateInfo _animateInfo = widget.controller.value;

  late final Tween<Offset> _offsetTween =
      Tween<Offset>(begin: Offset.zero, end: Offset.zero);
  late Animation<double> _fadeAnimation;
  late Animation<double> _offsetAnimation;
  @override
  void initState() {
    super.initState();
    _offsetAnimation = CurvedAnimation(
      parent: _controller,
      curve: Curves.ease,
    )..addStatusListener(onStatusChange);
    _fadeAnimation = Tween<double>(begin: 1, end: 0).animate(CurvedAnimation(
      parent: _controller,
      curve: Curves.ease,
    ))
      ..addStatusListener(onStatusChange);
    widget.controller.addListener(() {
      if (mounted) {
        setState(() {
          _animateInfo = widget.controller.value;

          _offsetTween.begin = _animateInfo.begin;
          _offsetTween.end = _animateInfo.end;
          _controller.reset();
          _controller.forward();
        });
      }
    });
  }

  void onAnimateEnd() {
    widget.gridController
        ._onAnimateEnd(_animateInfo.gen, widget.controller.index);
  }

  @override
  void dispose() {
    onAnimateEnd();
    _sizeNotifier.dispose();
    _controller.dispose();
    super.dispose();
  }

  void onStatusChange(AnimationStatus status) {
    if (status == AnimationStatus.completed) {
      onAnimateEnd();
    }
  }

  @override
  Widget build(BuildContext context) {
    WidgetsBinding.instance!.addPostFrameCallback((t) {
      if (!mounted) {
        return;
      }
      final RenderBox referenceBox = context.findRenderObject()! as RenderBox;
      final position = referenceBox.localToGlobal(Offset.zero);
      widget.controller.position = position;
      _sizeNotifier.value = referenceBox.size;
    });
    final Widget child =
        widget.builder(context, widget.item, widget.controller.index);

    final Widget transition = _animateInfo.type == _AnimateType.slide
        ? AnimatedBuilder(
            animation: _controller,
            child: child,
            builder: (BuildContext context, Widget? child) {
              return Transform.translate(
                  offset: _offsetTween.evaluate(_offsetAnimation),
                  child: child);
            })
        : FadeTransition(
            opacity: _fadeAnimation,
            child: child,
          );

    if (!widget.draggableConfiguration
        .enableChecker(widget.item, widget.controller.index)) {
      return transition;
    }

    void _onAccept() {
      if (widget.gridController._animateItems != null) {
        widget.onAccept?.call(widget.gridController._animateItems!);
      }
    }

    void _onDragComplete(int index) {
      widget.onDragCompleted?.call(widget.gridController._animateItems, index);
    }

    return DragTarget<int>(
      onWillAccept: (data) {
        widget.gridController._drag(data!, widget.controller.index);
        return true;
      },
      onAccept: (data) {
        widget.controller.accepted = true;
        _onAccept();
      },
      builder: (context, candidateData, rejectedData) {
        return LongPressDraggable(
          delay: widget.draggableConfiguration.longPressDelay,
          childWhenDragging:
              widget.draggableConfiguration.childWhenDraggingBuilder?.call(
                      context, widget.item, widget.controller.index, child) ??
                  Container(),
          feedback: ValueListenableBuilder<Size?>(
            builder: (context, size, _child) {
              return Material(
                color: Colors.transparent,
                child: widget.draggableConfiguration.feedbackBuilder?.call(
                        context,
                        widget.item,
                        widget.controller.index,
                        child,
                        size) ??
                    SizedBox(
                      width: size?.width,
                      height: size?.height,
                      child: child,
                    ),
              );
            },
            valueListenable: _sizeNotifier,
          ),
          data: widget.controller.currentIndex,
          onDragStarted: () {
            widget.draggableConfiguration.onDragStart
                ?.call(widget.controller.index);
          },
          onDragCompleted: () {
            if (!mounted) {
              return;
            }
            final bool accepted = widget.controller.accepted;
            if (!accepted) {
              _onDragComplete(widget.controller.index);
            }
          },
          onDraggableCanceled: (velocity, offset) {
            if (!mounted) {
              return;
            }
            _onAccept();
          },
          onDragEnd: (d) {
            if (!mounted) {
              return;
            }
            widget.draggableConfiguration.onDragEnd
                ?.call(widget.controller.index);
          },
          child: transition,
        );
      },
    );
  }
}

class _DraggableController extends ValueNotifier<_AnimateInfo> {
  /// grid item index
  final int index;

  final Key key = UniqueKey();
  // used to hold grid item's pixel position
  Offset? position;

  int? _currentIndex;

  /// current index after drag
  int get currentIndex => _currentIndex ?? index;

  bool accepted = false;

  _DraggableController(
    this.index,
  ) : super(const _AnimateInfo.slide(Offset.zero, Offset.zero, 0));
}

@immutable
class _AnimateInfo {
  final Offset begin;
  final Offset end;
  final int gen;
  final _AnimateType type;

  @override
  bool operator ==(Object o) =>
      o is _AnimateInfo &&
      begin == o.begin &&
      end == o.end &&
      gen == o.gen &&
      type == type;
  @override
  int get hashCode => Object.hash(begin, end, gen, type);

  const _AnimateInfo.slide(this.begin, this.end, this.gen)
      : type = _AnimateType.slide;

  const _AnimateInfo.fadeIn(this.gen)
      : begin = Offset.zero,
        end = Offset.zero,
        type = _AnimateType.fadeIn;
}

enum _AnimateType {
  fadeIn,
  slide,
}

class _QuietlySizeNotifier extends ValueNotifier<Size?> {
  _QuietlySizeNotifier(Size? value) : super(value);
  bool _disposed = false;
  @override
  void dispose() {
    _disposed = true;
    super.dispose();
  }

  @override
  void removeListener(VoidCallback listener) {
    if (!_disposed) {
      super.removeListener(listener);
    }
  }

  @override
  void addListener(VoidCallback listener) {
    if (!_disposed) {
      super.addListener(listener);
    }
  }

  @override
  void notifyListeners() {
    if (!_disposed) {
      super.notifyListeners();
    }
  }
}

class DraggableConfiguration<T> {
  final bool Function(T item, int index) enableChecker;
  final Duration longPressDelay;
  final Widget Function(BuildContext context, T item, int index, Widget child)?
      childWhenDraggingBuilder;
  final Widget Function(
          BuildContext context, T item, int index, Widget child, Size? size)?
      feedbackBuilder;
  final ValueChanged<int>? onDragStart;
  final ValueChanged<int>? onDragEnd;

  static bool _enableAll(dynamic item, int index) {
    return true;
  }

  static bool _disableAll(dynamic item, int index) {
    return false;
  }

  DraggableConfiguration({
    this.enableChecker = _enableAll,
    this.childWhenDraggingBuilder,
    this.feedbackBuilder,
    this.longPressDelay = kLongPressTimeout,
    this.onDragStart,
    this.onDragEnd,
  });

  const DraggableConfiguration.disabled()
      : enableChecker = _disableAll,
        longPressDelay = kLongPressTimeout,
        childWhenDraggingBuilder = null,
        feedbackBuilder = null,
        onDragStart = null,
        onDragEnd = null;
}
