import 'dart:async';
import 'dart:math';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:forme/forme.dart';
import 'package:image_picker/image_picker.dart';

import 'draggable_grid_view.dart';

typedef ImagePickerBuilder = Widget Function(
  bool supportCamera,
  ValueChanged<ImageSource> pick,
);

typedef GridItemBuilder = Widget Function(
  ValueChanged<int> remove,
  FormeImage item,
  int index,
  bool readOnly,
  bool enable,
);

typedef OnDragCompleted = void Function(
  FormeImagePickerFieldController field,
  int index,
);

class FormeImagePicker extends FormeField<List<FormeImage>> {
  final int? maximum;
  final DraggableConfiguration<FormeImage>? draggableConfiguration;

  /// called when GridItem is dropped and accepted by other **[DragTarget]**
  ///
  /// you **must** call [FormeImagePickerFieldController.remove] or FormeImagePickerFieldController.value to rebuild widget
  final OnDragCompleted? onDragCompleted;

  /// color of remove icon default is [Colors.redAccent]
  final Color? gridItemRemoveIconColor;

  /// padding of grid item , default is top:12 right:12
  final EdgeInsetsGeometry? gridItemPadding;

  final Color? imagePickerColor;
  final Color? imagePickerDisabledColor;
  final Widget? imagePickerChild;

  /// whether show image picker when readOnly or disabled, default is true
  final bool showImagePickerWhenReadOnly;

  final bool showGridItemRemoveIcon;
  final IconData? gridItemRemoveIcon;

  /// grid item remove icon size , default is 24
  final double gridItemRemoveIconSize;

  /// whether support capture with camera , default is false
  final bool supportCamera;
  final CameraDevice? preferredCameraDevice;

  final Widget? pickImageFromCameraOnBottomSheet;
  final Widget? pickImageFromGalleryOnBottomSheet;
  final Widget? cancelOnBottomSheet;

  final bool Function(XFile file)? fileTest;
  FormeImagePicker({
    Duration? animateDuration,
    this.maximum,
    this.draggableConfiguration,
    ImagePickerBuilder? pickerBuilder,
    GridItemBuilder? gridItemBuilder,
    required SliverGridDelegate gridDelegate,
    List<FormeImage>? initialValue,
    required String name,
    bool readOnly = false,
    Key? key,
    InputDecoration? decoration,
    FormeFieldDecorator<List<FormeImage>>? decorator,
    int? order,
    bool quietlyValidate = false,
    Duration? asyncValidatorDebounce,
    AutovalidateMode? autovalidateMode,
    FormeValueChanged<List<FormeImage>>? onValueChanged,
    FormeFocusChanged<List<FormeImage>>? onFocusChanged,
    FormeFieldValidationChanged<List<FormeImage>>? onValidationChanged,
    FormeFieldInitialed<List<FormeImage>>? onInitialed,
    FormeFieldSetter<List<FormeImage>>? onSaved,
    FormeValidator<List<FormeImage>>? validator,
    FormeAsyncValidator<List<FormeImage>>? asyncValidator,
    bool requestFocusOnUserInteraction = true,
    bool registrable = true,
    bool enabled = true,
    this.onDragCompleted,
    this.gridItemRemoveIconColor,
    this.gridItemPadding,
    this.showImagePickerWhenReadOnly = true,
    this.imagePickerColor,
    this.imagePickerDisabledColor,
    this.imagePickerChild,
    this.gridItemRemoveIcon,
    this.showGridItemRemoveIcon = true,
    this.gridItemRemoveIconSize = 24,
    this.fileTest,
    this.supportCamera = false,
    this.preferredCameraDevice,
    this.pickImageFromCameraOnBottomSheet,
    this.pickImageFromGalleryOnBottomSheet,
    this.cancelOnBottomSheet,
  }) : super(
            enabled: enabled,
            registrable: registrable,
            requestFocusOnUserInteraction: requestFocusOnUserInteraction,
            order: order,
            quietlyValidate: quietlyValidate,
            asyncValidatorDebounce: asyncValidatorDebounce,
            autovalidateMode: autovalidateMode,
            onValueChanged: onValueChanged,
            onFocusChanged: onFocusChanged,
            onValidationChanged: onValidationChanged,
            onInitialed: onInitialed,
            onSaved: onSaved,
            validator: validator,
            asyncValidator: asyncValidator,
            key: key,
            decorator: decorator ??
                (decoration == null
                    ? null
                    : FormeInputDecoratorBuilder(decoration: decoration)),
            readOnly: readOnly,
            name: name,
            initialValue: initialValue ?? <FormeImage>[],
            builder: (genericState) {
              final _FormeImagePickerState state =
                  genericState as _FormeImagePickerState;
              return DraggableGridView<_Item<FormeImage>>(
                animateDuration:
                    animateDuration ?? const Duration(milliseconds: 150),
                gridDelegate: gridDelegate,
                onAccept: state._onAccept,
                onDragCompleted: state._onDragCompleted,
                builder: (context, item, index) {
                  if (maximum != null &&
                      state.gridController.value.length - 1 >= maximum &&
                      item.value == null) {
                    return Container();
                  }
                  Widget child;
                  if (item.value == null) {
                    child = (pickerBuilder ?? state._defaultImagePicker)
                        .call(supportCamera, state._pickImage);
                  } else {
                    child = (gridItemBuilder ?? state._defaultGridItem).call(
                      (int index) {
                        if (readOnly) {
                          return;
                        }
                        state._remove(index);
                      },
                      item.value!,
                      index,
                      state.readOnly,
                      state.enabled,
                    );
                  }
                  return child;
                },
                draggableConfiguration: state._draggableConfiguration(),
                controller: state.gridController,
              );
            });

  @override
  FormeFieldState<List<FormeImage>> createState() => _FormeImagePickerState();
}

class _FormeImagePickerState extends FormeFieldState<List<FormeImage>> {
  @override
  FormeImagePicker get widget => super.widget as FormeImagePicker;

  late GridController<_Item<FormeImage>> gridController;

  late final ValueNotifier<bool> draggingNotifer =
      FormeMountedValueNotifier(false, this);

  @override
  void beforeInitiation() {
    super.beforeInitiation();
    gridController = GridController(_convert(List.of(initialValue)));
  }

  @override
  void updateFieldValueInDidUpdateWidget(
      FormeField<List<FormeImage>> oldWidget) {
    if (widget.maximum != null) {
      setValue(List.of(value).sublist(0, min(widget.maximum!, value.length)));
    }
  }

  List<_Item<FormeImage>> _convert(List<FormeImage> images) {
    List<_Item<FormeImage>> items = images.map((e) => _Item(e)).toList();
    if (widget.maximum != null && items.length > widget.maximum!) {
      items = items.sublist(0, widget.maximum);
    }

    final bool showImagePicker =
        (!readOnly && enabled) || widget.showImagePickerWhenReadOnly;
    if (showImagePicker &&
        (widget.maximum == null || items.length < widget.maximum!)) {
      items.add(_Item.empty());
    }
    return items;
  }

  @override
  void afterInitiation() {
    super.afterInitiation();
    gridController.addListener(() {
      didChange(gridController.value
          .where((element) => element.value != null)
          .map((e) => e.value!)
          .toList());
    });
  }

  @override
  void dispose() {
    draggingNotifer.dispose();
    gridController.dispose();
    super.dispose();
  }

  @override
  void onValueChanged(List<FormeImage> newValue) {
    gridController.value = _convert(newValue);
  }

  /// this method will decode image first rather than directly return [ImageProvider]
  Future<ImageProvider> _load(FormeImage file,
      [ImageConfiguration? configuration]) async {
    final ImageProvider image = await file.image;
    final Completer<ImageProvider> completer = Completer();
    final ImageStream stream =
        image.resolve(configuration ?? const ImageConfiguration());
    final ImageStreamListener listener =
        ImageStreamListener((info, synchronousCall) {
      file._broken = false;
      file._cache = image;
      completer.complete(image);
    }, onError: (error, trace) {
      file._broken = true;
      file._cache = null;
      completer.completeError(error, trace);
    });
    stream.addListener(listener);
    final Future<ImageProvider> future = completer.future;
    return future.whenComplete(() {
      //https://github.com/Sub6Resources/flutter_html/issues/780
      stream.removeListener(listener);
    });
  }

  void _remove(int index) {
    if (!mounted) {
      return;
    }
    gridController.removeAnimate([index], (items) {
      gridController.value = _convert(items
          .where((element) => element.value != null)
          .map((e) => e.value!)
          .toList());
    });
  }

  Widget _defaultGridItem(ValueChanged<int> remove, FormeImage item, int index,
      bool readOnly, bool enable) {
    Widget image;
    if (item.isBroken == false) {
      image = Image(
        image: item.cache!,
        fit: BoxFit.cover,
        width: double.infinity,
        height: double.infinity,
      );
    } else {
      image = FutureBuilder<ImageProvider>(
        future: _load(item),
        builder: (context, builder) {
          if (builder.hasData) {
            return Image(
              image: builder.data!,
              fit: BoxFit.cover,
              width: double.infinity,
              height: double.infinity,
            );
          }

          if (builder.hasError) {
            return const Center(
                child: Icon(
              Icons.broken_image,
              size: 36,
            ));
          }
          return const SizedBox.shrink();
        },
      );
    }

    return Stack(
      children: [
        Padding(
          padding: widget.gridItemPadding ??
              (widget.showGridItemRemoveIcon
                  ? EdgeInsets.only(
                      top: readOnly ? 0 : widget.gridItemRemoveIconSize / 2,
                      right: readOnly ? 0 : widget.gridItemRemoveIconSize / 2,
                    )
                  : EdgeInsets.zero),
          child: image,
        ),
        if (!readOnly && widget.showGridItemRemoveIcon)
          Positioned(
            top: 0,
            right: 0,
            child: InkWell(
              highlightColor: Colors.transparent,
              splashFactory: NoSplash.splashFactory,
              hoverColor: Colors.transparent,
              focusColor: Colors.transparent,
              child: Icon(
                widget.gridItemRemoveIcon ?? Icons.cancel,
                color: widget.gridItemRemoveIconColor ?? Colors.redAccent,
                size: widget.gridItemRemoveIconSize,
              ),
              onTap: () {
                remove(index);
              },
            ),
          )
      ],
    );
  }

  Future _pickImage(ImageSource source) async {
    if (source == ImageSource.camera && !widget.supportCamera) {
      return;
    }
    final ImagePicker _picker = ImagePicker();
    Iterable<XFile> images;
    if (widget.maximum == 1 ||
        value.length + 1 == widget.maximum ||
        source == ImageSource.camera) {
      //can select only one file
      final XFile? image = await _picker.pickImage(
        source: source,
        preferredCameraDevice:
            widget.preferredCameraDevice ?? CameraDevice.rear,
      );
      images = image == null ? [] : [image];
    } else {
      images = await _picker.pickMultiImage() ?? [];
    }

    if (!mounted || readOnly) {
      return;
    }

    if (widget.fileTest != null) {
      images = images.where(widget.fileTest!);
    }

    if (images.isEmpty) {
      return;
    }

    final List<FormeImage> currentValue = gridController.value
        .where((element) => element.value != null)
        .map((e) => e.value!)
        .toList();
    final int num = currentValue.length;
    if (widget.maximum != null && num >= widget.maximum!) {
      return;
    }
    final Iterable<FormeImage> needInserts = images
        .map(FormeImage.fromXFile)
        .where((element) => !currentValue.contains(element))
        .toList();
    if (needInserts.isEmpty) {
      return;
    }
    final List<FormeImage> list = [];
    for (final FormeImage item in needInserts) {
      if (!list.contains(item)) {
        list.add(item);
      }
    }
    gridController.value = _convert(currentValue..addAll(list));
  }

  Widget _defaultImagePicker(
      bool supportCamera, ValueChanged<ImageSource> pick) {
    return InkWell(
      highlightColor: Colors.transparent,
      splashFactory: NoSplash.splashFactory,
      hoverColor: Colors.transparent,
      focusColor: Colors.transparent,
      onTap: readOnly
          ? null
          : () {
              if (widget.supportCamera) {
                showModalBottomSheet<dynamic>(
                  shape: RoundedRectangleBorder(
                    borderRadius: BorderRadius.circular(10.0),
                  ),
                  context: context,
                  builder: (context) {
                    return Wrap(
                      children: [
                        InkWell(
                          onTap: () {
                            pick(ImageSource.camera);
                            Navigator.pop(context);
                          },
                          child: widget.pickImageFromCameraOnBottomSheet ??
                              const ListTile(
                                title: Center(
                                    child: Text('Pick Image fromr Camera')),
                              ),
                        ),
                        const Divider(),
                        InkWell(
                          onTap: () {
                            pick(ImageSource.gallery);
                            Navigator.pop(context);
                          },
                          child: widget.pickImageFromGalleryOnBottomSheet ??
                              const ListTile(
                                title: Center(
                                    child: Text('Pick Image fromr Gallery')),
                              ),
                        ),
                        Divider(
                          color: Colors.grey.withOpacity(0.1),
                          thickness: 10,
                        ),
                        InkWell(
                          onTap: () {
                            Navigator.pop(context);
                          },
                          child: widget.cancelOnBottomSheet ??
                              const ListTile(
                                title: Center(
                                  child: Text('Cancel'),
                                ),
                              ),
                        ),
                      ],
                    );
                  },
                );
              } else {
                pick(ImageSource.gallery);
              }
            },
      child: Padding(
        padding: widget.gridItemPadding ??
            (widget.showGridItemRemoveIcon
                ? EdgeInsets.only(
                    top: widget.gridItemRemoveIconSize / 2,
                    right: widget.gridItemRemoveIconSize / 2,
                  )
                : EdgeInsets.zero),
        child: Container(
          color: readOnly
              ? widget.imagePickerDisabledColor ??
                  Theme.of(context).disabledColor
              : widget.imagePickerColor ?? Colors.grey.withOpacity(0.3),
          child:
              widget.imagePickerChild ?? const Center(child: Icon(Icons.add)),
        ),
      ),
    );
  }

  DraggableConfiguration<_Item<FormeImage>> _draggableConfiguration() {
    if (widget.draggableConfiguration == null || readOnly) {
      return const DraggableConfiguration.disabled();
    }
    return DraggableConfiguration<_Item<FormeImage>>(
      enableChecker: (item, index) {
        if (item.value == null) {
          return false;
        }
        return widget.draggableConfiguration!.enableChecker(item.value!, index);
      },
      longPressDelay: widget.draggableConfiguration!.longPressDelay,
      childWhenDraggingBuilder:
          widget.draggableConfiguration!.childWhenDraggingBuilder == null
              ? null
              : (context, item, index, child) {
                  return widget
                          .draggableConfiguration!.childWhenDraggingBuilder!(
                      context, item.value!, index, child);
                },
      feedbackBuilder: widget.draggableConfiguration!.feedbackBuilder == null
          ? null
          : (context, item, index, child, size) {
              return widget.draggableConfiguration!.feedbackBuilder!(
                  context, item.value!, index, child, size);
            },
      onDragStart: (int index) {
        draggingNotifer.value = true;
        widget.draggableConfiguration!.onDragStart?.call(index);
      },
      onDragEnd: (int index) {
        draggingNotifer.value = false;
        widget.draggableConfiguration!.onDragEnd?.call(index);
      },
    );
  }

  void _onAccept(List<_Item<FormeImage>> value) {
    gridController.value = value;
  }

  void _onDragCompleted(List<_Item<FormeImage>>? value, int index) {
    if (widget.onDragCompleted != null) {
      widget.onDragCompleted!
          .call(controller as FormeImagePickerFieldController, index);
    } else if (value != null) {
      gridController.value = value;
    }
  }

  @override
  FormeFieldController<List<FormeImage>> createFormeFieldController() {
    return FormeImagePickerFieldController._(
      super.createFormeFieldController(),
      this,
    );
  }
}

class FormeImagePickerFieldController
    extends FormeFieldControllerDelegate<List<FormeImage>> {
  final _FormeImagePickerState _state;
  final ValueListenable<bool> draggingListenable;
  FormeImagePickerFieldController._(
      FormeFieldController<List<FormeImage>> delegate, this._state)
      : draggingListenable =
            FormeValueListenableDelegate(_state.draggingNotifer),
        super(delegate);

  /// animate & remove item at index
  void remove(int index) {
    final int max = _state.gridController.value
        .where((element) => element.value != null)
        .length;
    if (index < 0 || index >= max) {
      return;
    }
    _state._remove(index);
  }
}

abstract class FormeImage {
  Future<ImageProvider> get image;
  bool? _broken;
  ImageProvider? _cache;

  ///  typically this `image` is not a real image or can not be decoded by current platform
  ///
  /// return null if image not decoded
  bool? get isBroken => _broken;

  /// cached imageprovider after successful load from [image]
  ///
  /// will not be null if [isBroken] == false
  ImageProvider? get cache => _cache;

  /// get file name
  String? get name;

  /// get image bytes,used to save to a file
  ///
  /// there's no need to implementing this method in your custom `FormeImage`
  Future<Uint8List?> readAsBytes();

  static FormeImage fromXFile(XFile xFile) {
    return _FormeXFile(xFile);
  }

  static FormeImage fromNetwork(String url) {
    return _NetworkImage(url);
  }
}

class _NetworkImage extends FormeImage {
  final String url;

  _NetworkImage(this.url);

  @override
  Future<ImageProvider> get image async {
    return Image.network(url).image;
  }

  @override
  String? get name => throw UnimplementedError();

  @override
  Future<Uint8List?> readAsBytes() {
    throw UnimplementedError();
  }
}

class _FormeXFile extends FormeImage {
  final XFile xFile;

  _FormeXFile(this.xFile);
  @override
  Future<ImageProvider> get image async {
    return Image.memory(await readAsBytes()).image;
  }

  /// [ImagePicker] does not support get original file name in some os,but in web it does
  @override
  String? get name => xFile.name;

  @override
  Future<Uint8List> readAsBytes() async => xFile.readAsBytes();
}

class _Item<T> {
  final T? value;

  const _Item(this.value);

  _Item.empty() : value = null;
}
